from polars import Expr

from polars_ta.wq.arithmetic import log1p, expm1  # noqa
# log return
# 对数收益
from polars_ta.wq.time_series import ts_log_diff as ts_log_return  # noqa
# simple percentage return
# 简单收益
from polars_ta.wq.time_series import ts_returns as ts_percent_return  # noqa


def ts_cum_return(close: Expr) -> Expr:
    """
    cumulative return
    累计收益
    """

    return close / close.drop_nulls().first()


def simple_to_log_return(x: Expr) -> Expr:
    """简单收益率 转 对数收益率"""
    return log1p(x)


def log_to_simple_return(x: Expr) -> Expr:
    """对数收益率 转 简单收益率"""
    return expm1(x)



from polars_ta.ta.momentum import *  # noqa
from polars_ta.ta.operators import *  # noqa
from polars_ta.ta.overlap import *  # noqa
from polars_ta.ta.price import *  # noqa
from polars_ta.ta.statistic import *  # noqa
from polars_ta.ta.transform import *  # noqa
from polars_ta.ta.volatility import *  # noqa
from polars_ta.ta.volume import *  # noqa



from polars import Expr

from polars_ta import TA_EPSILON
from polars_ta.ta.momentum import RSI  # noqa
from polars_ta.ta.momentum import RSV
from polars_ta.ta.price import TYPPRICE
from polars_ta.tdx.choice import IF
from polars_ta.tdx.reference import DIFF
from polars_ta.tdx.reference import MA
from polars_ta.tdx.reference import REF
from polars_ta.tdx.reference import SMA_CN
from polars_ta.tdx.reference import SUM
from polars_ta.tdx.reference import TR
from polars_ta.tdx.statistic import AVEDEV


def ATR(HIGH: Expr, LOW: Expr, CLOSE: Expr, N: int = 14) -> Expr:
    """

    Notes
    -----
    Warning: it is different with talib.ATR
    与talib.ATR不同

    """
    return MA(TR(HIGH, LOW, CLOSE), N)


def BIAS(CLOSE: Expr, N: int = 6) -> Expr:
    """
    BIAS1 :(CLOSE-MA(CLOSE,N1))/MA(CLOSE,N1)*100;
    BIAS2 :(CLOSE-MA(CLOSE,N2))/MA(CLOSE,N2)*100;
    BIAS3 :(CLOSE-MA(CLOSE,N3))/MA(CLOSE,N3)*100;

    """
    return CLOSE / MA(CLOSE, N) - 1  # * 100


def CCI(HIGH: Expr, LOW: Expr, CLOSE: Expr, N: int = 14) -> Expr:
    """
    TYP:=(HIGH+LOW+CLOSE)/3;
    CCI:(TYP-MA(TYP,N))*1000/(15*AVEDEV(TYP,N));

    Notes
    -----
    Avoid using `AVEDEV`, it is slow
    AVEDEV计算慢，少用

    """
    TYP = TYPPRICE(HIGH, LOW, CLOSE)
    return (TYP - MA(TYP, N)) / (0.015 * AVEDEV(TYP, N))


def KDJ(HIGH: Expr, LOW: Expr, CLOSE: Expr, N: int = 9, M1: int = 3, M2: int = 3) -> Expr:
    """
    RSV:=(CLOSE-LLV(LOW,N))/(HHV(HIGH,N)-LLV(LOW,N))*100;
    K:SMA(RSV,M1,1);
    D:SMA(K,M2,1);
    J:3*K-2*D;

    """
    rsv = RSV(HIGH, LOW, CLOSE, N)
    k = SMA_CN(rsv, M1, 1)
    d = SMA_CN(k, M2, 1)
    j = k * 3 - d * 2
    # return k, d, j
    return j


def MTM(CLOSE: Expr, N: int = 12) -> Expr:
    """
    MTM:CLOSE-REF(CLOSE,MIN(BARSCOUNT(C),N));
    MTMMA:MA(MTM,M);
    """
    # return CLOSE - REF(CLOSE, N)
    return DIFF(CLOSE, N)


def MFI(CLOSE: Expr, HIGH: Expr, LOW: Expr, VOL: Expr, N: int = 14) -> Expr:
    """

    TYP := (HIGH + LOW + CLOSE)/3;
    V1:=SUM(IF(TYP>REF(TYP,1),TYP*VOL,0),N)/SUM(IF(TYP<REF(TYP,1),TYP*VOL,0),N);
    MFI:100-(100/(1+V1));

    """
    TYP = TYPPRICE(HIGH, LOW, CLOSE)
    LT = REF(TYP, 1)
    V1 = SUM(IF(TYP > LT, TYP * VOL, 0), N) / (SUM(IF(TYP < LT, TYP * VOL, 0), N) + TA_EPSILON)
    return (1 - (1 / (1 + V1)))  # * 100



from polars import Expr

from polars_ta.wq import ts_corr, ts_zscore, ts_std_dev, ts_regression_slope


def ts_RSRS_R2(high: Expr, low: Expr, n: int = 18, m: int = 600) -> Expr:
    """光大RSRS指标，中金QRS指标，R^2调整

    References
    ----------
    中金：金融工程视角下的技术择时艺术
    """
    a = ts_corr(high, low, n)
    return ts_zscore(ts_std_dev(high, n) / ts_std_dev(low, n) * a, m) * (a ** 2)


def ts_RSRS(high: Expr, low: Expr, n: int = 18, m: int = 600) -> Expr:
    """光大RSRS指标，中金QRS指标

    References
    ----------
    中金：金融工程视角下的技术择时艺术
    """
    return ts_zscore(ts_regression_slope(high, low, n), m)



"""
We follow the spirit of

https://github.com/pola-rs/polars/issues/9261

and adjusted it to fit our needs. The usage is as follows

expr.ta.func(..., skip_nan=False, output_idx=None, schema=None, schema_format='{}', nan_to_null=False)

skip_nan: bool
    reduces speed
output_idx: int
    select a single column when outputing multiple columns
schema: list or tuple
    assign column names for multiple output columns
schema_format: str
    assign data types for multiple output columns
nan_to_null: bool
    replace all nan to null when outputing



以下是在github issues中 @cmdlineluser 提供的 注册命名空间 方案

https://github.com/pola-rs/polars/issues/9261

本人做了一定的调整。使用方法如下

expr.ta.func(..., skip_nan=False, output_idx=None, schema=None, schema_format='{}', nan_to_null=False)

skip_nan: bool
    是否跳过空值。可以处理停牌无数据的问题，但会降低运行速度
output_idx: int
    多列输出时，选择只输出其中一列
schema: list or tuple
    返回为多列时，会组装成struct，可以提前设置子列的名字
schema_format: str
    为子列名指定格式
nan_to_null: bool
    返回值是否改成null

其它参数按**位置参数**和**命名参数**输入皆可
"""
import numpy as np
from polars import Expr, Float64, DataFrame, api
from polars import Series, Struct


def func_wrap_mn(func, cols,
                 *args,
                 skip_nan=False, output_idx=None, schema=None, schema_format='{}', nan_to_null=False,
                 **kwargs):
    """multiple input multiple output, compatible with single input single output
    多输入多输出，兼容一输入一输出

    Parameters
    ----------
    func
    cols
    args
    skip_nan
    output_idx
    schema
    schema_format
    nan_to_null
    kwargs

    Returns
    -------
    Series

    """
    if cols.dtype.base_type() == Struct:
        # struct(['A', 'B']).ta.AROON
        _cols = [cols.struct[field] for field in cols.dtype.to_schema()]
    else:
        # col('A').ta.BBANDS
        _cols = [cols]

    if skip_nan:
        _cols = [c.cast(Float64).to_numpy() for c in _cols]
        _cols = np.vstack(_cols)

        # move nan to head
        idx1 = (~np.isnan(_cols).any(axis=0)).argsort(kind='stable')
        # restore nan
        idx2 = idx1.argsort(kind='stable')

        _cols = [_cols[i, idx1] for i in range(len(_cols))]
        result = func(*_cols, *args, **kwargs)

        if isinstance(result, tuple):
            result = tuple([_[idx2] for _ in result])
        else:
            result = result[idx2]
    else:
        result = func(*_cols, *args, **kwargs)

    if isinstance(result, tuple):
        if output_idx is None:
            # struct(['A').ta.BBANDS
            # you need alias outside, finally unnest
            if schema is None:
                schema = [f'column_{i}' for i in range(len(result))]

            schema = [schema_format.format(name) for name in schema]
            # nan_to_null is not effective for struct
            # nan_to_null 对struct中的nan无效
            return DataFrame(result, schema=schema, nan_to_null=nan_to_null).to_struct('')
        # output only one column
        if 0 <= output_idx < len(result):
            return Series(result[output_idx], nan_to_null=nan_to_null)
    elif not isinstance(result, Series):
        # col('A').bn.move_rank
        return Series(result, nan_to_null=nan_to_null)
    else:
        # col('A').ta.COS
        if nan_to_null:
            return result.fill_nan(None)
        else:
            return result


def func_wrap_11(func, cols,
                 *args,
                 skip_nan=False, output_idx=None, schema=None, schema_format='{}', nan_to_null=False,
                 **kwargs):
    """single input single output is faster than multiple input multiple output
    一输入，一输出。处理速度能更快一些"""
    _cols = cols

    if skip_nan:
        _cols = _cols.cast(Float64).to_numpy()

        # move nan to head
        idx1 = (~np.isnan(_cols)).argsort(kind='stable')
        # restore nan
        idx2 = idx1.argsort(kind='stable')

        _cols = _cols[idx1]
        result = func(_cols, *args, **kwargs)

        result = result[idx2]
    else:
        result = func(_cols, *args, **kwargs)

    if isinstance(result, Series):
        if nan_to_null:
            return result.fill_nan(None)
        else:
            return result
    else:
        return Series(result, nan_to_null=nan_to_null)


class FuncHelper:
    def __init__(self, expr: Expr, lib=None, wrap=None) -> None:
        """initialization

        Parameters
        ----------
        expr
        lib:
            third-party packages required
        wrap:
            wrapper for third-party packages


        初始化

        Parameters
        ----------
        expr
        lib:
            需要调用的第三方库
        wrap:
            对第三方库的封装方法

        """
        object.__setattr__(self, '_expr', expr)
        object.__setattr__(self, '_lib', lib)
        object.__setattr__(self, '_wrap', wrap)

    def __getattribute__(self, name: str):
        _expr: Expr = object.__getattribute__(self, '_expr')
        _lib = object.__getattribute__(self, '_lib')
        _wrap = object.__getattribute__(self, '_wrap')
        _func = getattr(_lib, name)
        return (
            lambda *args, skip_nan=False, output_idx=None, schema=None, schema_format='{}', nan_to_null=False, **kwargs:
            _expr.map_batches(
                lambda cols: _wrap(_func, cols,
                                   *args,
                                   skip_nan=skip_nan, output_idx=output_idx, schema=schema, schema_format=schema_format, nan_to_null=nan_to_null,
                                   **kwargs)
            )
        )


@api.register_expr_namespace('ta')
class TaLibHelper(FuncHelper):
    def __init__(self, expr: Expr) -> None:
        import talib as ta
        super().__init__(expr, ta, func_wrap_mn)


@api.register_expr_namespace('bn')
class BottleneckHelper(FuncHelper):
    def __init__(self, expr: Expr) -> None:
        import bottleneck as bn
        super().__init__(expr, bn, func_wrap_11)



# this code is auto generated by tools/prefix_ta.py

from polars_ta.ta.momentum import APO as ts_APO  # noqa
from polars_ta.ta.momentum import AROON_aroondown as ts_AROON_aroondown  # noqa
from polars_ta.ta.momentum import AROON_aroonup as ts_AROON_aroonup  # noqa
from polars_ta.ta.momentum import MACD_macd as ts_MACD_macd  # noqa
from polars_ta.ta.momentum import MACD_macdhist as ts_MACD_macdhist  # noqa
from polars_ta.ta.momentum import MACD_macdsignal as ts_MACD_macdsignal  # noqa
from polars_ta.ta.momentum import MOM as ts_MOM  # noqa
from polars_ta.ta.momentum import PPO as ts_PPO  # noqa
from polars_ta.ta.momentum import ROC as ts_ROC  # noqa
from polars_ta.ta.momentum import ROCP as ts_ROCP  # noqa
from polars_ta.ta.momentum import ROCR as ts_ROCR  # noqa
from polars_ta.ta.momentum import ROCR100 as ts_ROCR100  # noqa
from polars_ta.ta.momentum import RSI as ts_RSI  # noqa
from polars_ta.ta.momentum import RSV as ts_RSV  # noqa
from polars_ta.ta.momentum import STOCHF_fastd as ts_STOCHF_fastd  # noqa
from polars_ta.ta.momentum import TRIX as ts_TRIX  # noqa
from polars_ta.ta.momentum import WILLR as ts_WILLR  # noqa
from polars_ta.ta.overlap import BBANDS_upperband as ts_BBANDS_upperband  # noqa
from polars_ta.ta.overlap import DEMA as ts_DEMA  # noqa
from polars_ta.ta.overlap import EMA as ts_EMA  # noqa
from polars_ta.ta.overlap import KAMA as ts_KAMA  # noqa
from polars_ta.ta.overlap import MIDPOINT as ts_MIDPOINT  # noqa
from polars_ta.ta.overlap import MIDPRICE as ts_MIDPRICE  # noqa
from polars_ta.ta.overlap import RMA as ts_RMA  # noqa
from polars_ta.ta.overlap import SMA as ts_SMA  # noqa
from polars_ta.ta.overlap import TEMA as ts_TEMA  # noqa
from polars_ta.ta.overlap import TRIMA as ts_TRIMA  # noqa
from polars_ta.ta.overlap import WMA as ts_WMA  # noqa
from polars_ta.ta.price import AVGPRICE  # noqa
from polars_ta.ta.price import MEDPRICE  # noqa
from polars_ta.ta.price import TYPPRICE  # noqa
from polars_ta.ta.price import WCLPRICE  # noqa
from polars_ta.ta.statistic import BETA as ts_BETA  # noqa
from polars_ta.ta.statistic import CORREL as ts_CORREL  # noqa
from polars_ta.ta.statistic import LINEARREG as ts_LINEARREG  # noqa
from polars_ta.ta.statistic import LINEARREG_ANGLE as ts_LINEARREG_ANGLE  # noqa
from polars_ta.ta.statistic import LINEARREG_INTERCEPT as ts_LINEARREG_INTERCEPT  # noqa
from polars_ta.ta.statistic import LINEARREG_SLOPE as ts_LINEARREG_SLOPE  # noqa
from polars_ta.ta.statistic import STDDEV as ts_STDDEV  # noqa
from polars_ta.ta.statistic import TSF as ts_TSF  # noqa
from polars_ta.ta.statistic import VAR as ts_VAR  # noqa
from polars_ta.ta.transform import ACOS  # noqa
from polars_ta.ta.transform import ASIN  # noqa
from polars_ta.ta.transform import ATAN  # noqa
from polars_ta.ta.transform import CEIL  # noqa
from polars_ta.ta.transform import COS  # noqa
from polars_ta.ta.transform import COSH  # noqa
from polars_ta.ta.transform import EXP  # noqa
from polars_ta.ta.transform import FLOOR  # noqa
from polars_ta.ta.transform import LN  # noqa
from polars_ta.ta.transform import LOG10  # noqa
from polars_ta.ta.transform import SIN  # noqa
from polars_ta.ta.transform import SINH  # noqa
from polars_ta.ta.transform import SQRT  # noqa
from polars_ta.ta.transform import TAN  # noqa
from polars_ta.ta.transform import TANH  # noqa
from polars_ta.ta.volatility import ATR as ts_ATR  # noqa
from polars_ta.ta.volatility import NATR as ts_NATR  # noqa
from polars_ta.ta.volatility import TRANGE as ts_TRANGE  # noqa
from polars_ta.ta.volume import AD as ts_AD  # noqa
from polars_ta.ta.volume import ADOSC as ts_ADOSC  # noqa
from polars_ta.ta.volume import OBV as ts_OBV  # noqa


from polars import Expr

from polars_ta.tdx.reference import MA
from polars_ta.tdx.statistic import STDP


def BOLL(close: Expr, M: int = 20, N: int = 2) -> Expr:
    """
    BOLL:MA(CLOSE,M);
    UB:BOLL+2*STD(CLOSE,M);
    LB:BOLL-2*STD(CLOSE,M);
    """
    ma = MA(close, M)
    # it should be total standard deviation, the value is smaller than sample standard deviation.
    # 这里是总体标准差，值比样本标准差小。部分软件使用样本标准差是错误的，
    _std = STDP(close, M)
    UB = ma + _std * N
    return UB



from polars import Expr


def AVGPRICE(open: Expr, high: Expr, low: Expr, close: Expr) -> Expr:
    return (open + high + low + close) / 4


def MEDPRICE(high: Expr, low: Expr) -> Expr:
    return (high + low) / 2


def TYPPRICE(high: Expr, low: Expr, close: Expr) -> Expr:
    return (high + low + close) / 3


def WCLPRICE(high: Expr, low: Expr, close: Expr) -> Expr:
    return (high + low + close * 2) / 4



from polars import Boolean, Int32, UInt16
from polars import Expr
from polars import when

from polars_ta.ta.overlap import EMA as _ema
from polars_ta.ta.overlap import SMA as MA
from polars_ta.ta.volatility import TRANGE as TR  # noqa
from polars_ta.tdx._nb import roll_bars_since_n
from polars_ta.utils.numba_ import batches_i1_o1
from polars_ta.utils.pandas_ import roll_rank
from polars_ta.wq.arithmetic import max_ as MAX  # noqa
from polars_ta.wq.arithmetic import min_ as MIN  # noqa
from polars_ta.wq.time_series import ts_arg_max as HHVBARS  # noqa
from polars_ta.wq.time_series import ts_arg_min as LLVBARS  # noqa
from polars_ta.wq.time_series import ts_count as COUNT  # noqa
from polars_ta.wq.time_series import ts_decay_linear as WMA  # noqa
from polars_ta.wq.time_series import ts_delay as REF  # noqa
from polars_ta.wq.time_series import ts_delta as DIFF  # noqa
from polars_ta.wq.time_series import ts_max as HHV  # noqa
from polars_ta.wq.time_series import ts_min as LLV  # noqa
from polars_ta.wq.time_series import ts_product as MULAR  # noqa
from polars_ta.wq.time_series import ts_sum as SUM


def BARSLAST(condition: Expr) -> Expr:
    """# of Observations since last time condition was true
    上一次X不为0到现在的天数"""
    a = condition.cum_count()
    b = when(condition.cast(Boolean)).then(a).otherwise(None).forward_fill()
    return a - b


def BARSLASTCOUNT(condition: Expr) -> Expr:
    """Cumulative count of continuous true observations
    统计连续满足条件的周期数"""
    a = condition.cast(Int32).cum_sum()
    b = when(~condition.cast(Boolean)).then(a).otherwise(None).forward_fill().fill_null(0)
    return a - b


def BARSSINCE(condition: Expr) -> Expr:
    """# of observations since the first time condition was true
    第一次X不为0到现在的天数"""
    a = condition.cum_count()
    b = condition.cast(Boolean).arg_true().first()
    return a - b


def BARSSINCEN(condition: Expr, N: int = 30) -> Expr:
    """# of Observations since the first time condition was true (rolling within N observations)
    N周期内第一次X不为0到现在的天数"""
    return condition.cast(Boolean).map_batches(lambda x1: batches_i1_o1(x1.to_numpy(), roll_bars_since_n, N, dtype=UInt16))


def CUMSUM(close: Expr) -> Expr:
    """SUM(close, 0)"""
    return close.cum_sum()


def DMA(close: Expr, alpha: float = 0.5) -> Expr:
    """DMA(X,alpha), (Exponential moving average given alpha)
    Y = alpha * X + (1 - alpha) * last_Y
    requires 0 < alpha < 1

    求X的动态移动平均.
    算法:Y=A*X+(1-A)*Y',其中Y'表示上一周期Y值,A必须大于0且小于1.A支持变量"""
    return close.ewm_mean(alpha=alpha, adjust=False, min_periods=1)


def EMA(close: Expr, N: int = 30) -> Expr:
    """EMA(X,N): Exponential moving average given N

    Y = X * 2/(N+1) + last_Y * (N-1)/(N+1)

    X的N日指数移动平均.算法:Y=(X*2+Y'*(N-1))/(N+1)
    EMA(X,N)相当于SMA(X,N+1,2),N支持变量"""
    return _ema(close, N)


def EXPMA(close: Expr, N: int = 30) -> Expr:
    return _ema(close, N)


def EXPMEMA(close: Expr, N: int = 30) -> Expr:
    """Slow version of EMA. Do not use it unless you have to
    EXPMEMA(X,M),X的M日指数平滑移动平均。EXPMEMA同EMA(即EXPMA)的差别在于他的起始值为一平滑值

    Notes
    -----
    等价于talib.EMA，由于比EMA慢，少用

    """
    sma = MA(close, N)
    x = when(close.cum_count() < N).then(sma).otherwise(close)
    return x.ewm_mean(span=N, adjust=False, min_periods=1)


def HOD(close: Expr, N: int = 30) -> Expr:
    """rolling rank of each data in descending order
    HOD(X,N):求当前X数据是N周期内的第几个高值,N=0则从第一个有效值开始"""
    return close.map_batches(lambda a: roll_rank(a, N, pct=False, ascending=False))


def LOD(close: Expr, N: int = 30) -> Expr:
    """rolling rank of each data in ascending order
    LOD(X,N):求当前X数据是N周期内的第几个低值"""
    return close.map_batches(lambda a: roll_rank(a, N, pct=False, ascending=True))


def MEMA(close: Expr, N: int = 30) -> Expr:
    """Exponential moving average given N
    Y = X / N + last_Y * (N-1) / N

    MEMA(X,N):X的N日平滑移动平均,如Y=(X+Y'*(N-1))/N
 MEMA(X,N)相当于SMA(X,N,1)"""
    raise


def RANGE(a: Expr, b: Expr, c: Expr) -> Expr:
    """A在B和C范围之间,B<A<C."""
    return (b < a) & (a < c)


def SMA_CN(X: Expr, N: int, M: int) -> Expr:
    """Exponential Moving average given alpha = M/N
    Y = X * M/N + last_Y * (N-M)/N

    用法:SMA(X,N,M),X的N日移动平均,M为权重,若Y=SMA(X,N,M)则Y=(X*M+Y'*(N-M))/N

    !!!为防止与talib版SMA误用，这里去了默认值1
    """
    return X.ewm_mean(alpha=M / N, adjust=False, min_periods=1)


def SUMIF(condition: Expr, close: Expr, N: int = 30) -> Expr:
    return SUM(condition.cast(Int32) * close, N)


def TMA(close: Expr, N: int = 30) -> Expr:
    """TMA(X,A,B),A和B必须小于1,算法	Y=(A*Y'+B*X),其中Y'表示上一周期Y值.初值为X"""
    raise


def FILTER(close: Expr, N: int = 30) -> Expr:
    raise



from polars import Expr

from polars_ta import TA_EPSILON
from polars_ta.ta.operators import MAX
from polars_ta.ta.operators import MIN
from polars_ta.ta.overlap import EMA
from polars_ta.ta.overlap import RMA
from polars_ta.ta.overlap import SMA
from polars_ta.wq.arithmetic import max_
from polars_ta.wq.time_series import ts_delta, ts_arg_max, ts_arg_min
from polars_ta.wq.time_series import ts_returns


def APO(close: Expr, fastperiod: int = 12, slowperiod: int = 26, matype: int = 0) -> Expr:
    if matype == 0:
        return SMA(close, fastperiod) - SMA(close, slowperiod)
    else:
        return EMA(close, fastperiod) - EMA(close, slowperiod)


def AROON_aroondown(high: Expr, low: Expr, timeperiod: int = 14) -> Expr:
    """
    Lower band:
    下轨:(N-LLVBARS(L,N))/N*100,COLORGREEN;
    """
    return 1 - ts_arg_min(low, timeperiod, reverse=True) / timeperiod


def AROON_aroonup(high: Expr, low: Expr, timeperiod: int = 14) -> Expr:
    """
    Upper band:
    上轨:(N-HHVBARS(H,N))/N*100,COLORRED;

    Notes
    -----
    You cannot use pd.Series.rolling().arg_max() with reverse order, which leads to a larger result when there are two or more high points
    so we don't use pandas
    pd.Series.rolling().arg_max()没有逆序，导致出现两个及以上最高点时，结果偏大

    """
    return 1 - ts_arg_max(high, timeperiod, reverse=True) / timeperiod


def MACD_macd(close: Expr, fastperiod: int = 12, slowperiod: int = 26) -> Expr:
    """

    Notes
    -----
    When counting how many data we have
    we refer to `slowperiod`, while `talib.MACD` refers to `fastperiod`
    talib.MACD有效数据按fastperiod，而本项目按slowperiod

    """
    return EMA(close, fastperiod) - EMA(close, slowperiod)


def MACD_macdhist(close: Expr, fastperiod: int = 12, slowperiod: int = 26, signalperiod: int = 9) -> Expr:
    """

    Notes
    -----
    Chinese version is multiplied by 2
    中国版多了乘2

    """
    macd = MACD_macd(close, fastperiod, slowperiod)
    signal = EMA(macd, signalperiod)
    return macd - signal


def MACD_macdsignal(close: Expr, fastperiod: int = 12, slowperiod: int = 26, signalperiod: int = 9) -> Expr:
    macd = MACD_macd(close, fastperiod, slowperiod)
    signal = EMA(macd, signalperiod)
    return signal


def MOM(close: Expr, timeperiod: int = 10) -> Expr:
    return ts_delta(close, timeperiod)


def PPO(close: Expr, fastperiod: int = 12, slowperiod: int = 26, matype: int = 0) -> Expr:
    if matype == 0:
        return SMA(close, fastperiod) / SMA(close, slowperiod) - 1
    else:
        return EMA(close, fastperiod) / EMA(close, slowperiod) - 1


def ROC(close: Expr, timeperiod: int = 10) -> Expr:
    return ROCP(close, timeperiod) * 100


def ROCP(close: Expr, timeperiod: int = 10) -> Expr:
    return ts_returns(close, timeperiod)


def ROCR(close: Expr, timeperiod: int = 10) -> Expr:
    return close / close.shift(timeperiod)


def ROCR100(close: Expr, timeperiod: int = 10) -> Expr:
    return ROCR(close, timeperiod) * 100


def RSI(close: Expr, timeperiod: int = 14) -> Expr:
    dif = close.diff().fill_null(0)
    return RMA(max_(dif, 0), timeperiod) / (RMA(dif.abs(), timeperiod) + TA_EPSILON)  # * 100


def RSV(high: Expr, low: Expr, close: Expr, timeperiod: int = 5) -> Expr:
    """

    Notes
    -----
    Also called `STOCHF_fastk`, `talib.STOCHF` is multiplied by 100,
    and it is very similar to the `WILLR` indicator
    又名STOCHF_fastk, talib.STOCHF版相当于多乘了100，与WILLR指标又很像

    """
    a = MAX(high, timeperiod)
    b = MIN(low, timeperiod)

    return (close - b) / (a - b + TA_EPSILON)


def STOCHF_fastd(high: Expr, low: Expr, close: Expr, fastk_period: int = 5, fastd_period: int = 3) -> Expr:
    return SMA(RSV(high, low, close, fastk_period), fastd_period)


def TRIX(close: Expr, timeperiod: int = 30) -> Expr:
    EMA1 = EMA(close, timeperiod)
    EMA2 = EMA(EMA1, timeperiod)
    EMA3 = EMA(EMA2, timeperiod)
    return ROCP(EMA3, 1)


def WILLR(high: Expr, low: Expr, close: Expr, timeperiod: int = 14) -> Expr:
    """

    Notes
    -----
    `talib.WILLR` is multiplied by -100, but I think it is unnecessary
    talib.WILLR版相当于多乘了-100，但个人认为没有必要

    References
    ----------
    https://www.investopedia.com/terms/w/williamsr.asp
    https://school.stockcharts.com/doku.php?id=technical_indicators:williams_r

    """
    a = MAX(high, timeperiod)
    b = MIN(low, timeperiod)

    return (a - close) / (a - b + TA_EPSILON)



from polars import Expr

from polars_ta.wq.arithmetic import abs_ as ABS  # noqa
from polars_ta.wq.arithmetic import add as ADD  # noqa
from polars_ta.wq.arithmetic import arc_cos as ACOS  # noqa
from polars_ta.wq.arithmetic import arc_sin as ASIN  # noqa
from polars_ta.wq.arithmetic import arc_tan as ATAN  # noqa
from polars_ta.wq.arithmetic import ceiling as CEILING  # noqa
from polars_ta.wq.arithmetic import cos as COS  # noqa
from polars_ta.wq.arithmetic import exp as EXP  # noqa
from polars_ta.wq.arithmetic import floor as FLOOR  # noqa
from polars_ta.wq.arithmetic import fraction as FRACPART  # noqa
from polars_ta.wq.arithmetic import log as LN  # noqa # 自然对数 (log base e)
from polars_ta.wq.arithmetic import log10 as LOG  # noqa # 10为底的对数 (log base 10)
from polars_ta.wq.arithmetic import max_ as MAX  # noqa
from polars_ta.wq.arithmetic import min_ as MIN  # noqa
from polars_ta.wq.arithmetic import mod as MOD  # noqa
from polars_ta.wq.arithmetic import power as POW  # noqa
from polars_ta.wq.arithmetic import reverse as REVERSE  # noqa
from polars_ta.wq.arithmetic import round_ as _round  # noqa
from polars_ta.wq.arithmetic import sign as SIGN  # noqa
from polars_ta.wq.arithmetic import sin as SIN  # noqa
from polars_ta.wq.arithmetic import sqrt as SQRT  # noqa
from polars_ta.wq.arithmetic import subtract as SUB  # noqa
from polars_ta.wq.arithmetic import tan as TAN  # noqa
from polars_ta.wq.arithmetic import truncate as INTPART  # noqa

SGN = SIGN


def ROUND(x: Expr) -> Expr:
    """Round input to closest integer."""
    return _round(x, 0)


def ROUND2(x: Expr, decimals: int = 0) -> Expr:
    """Round input to closest integer."""
    return _round(x, decimals)


def BETWEEN(a: Expr, b: Expr, c: Expr) -> Expr:
    """1 if B<=A<=C or C<=A<=B else 0

    BETWEEN(A,B,C)表示A处于B和C之间时返回1(B<=A<=C或C<=A<=B),否则返回0"""
    x1 = (b <= a) & (a <= c)
    x2 = (c <= a) & (a <= b)
    return x1 | x2



# this code is auto generated by tools/prefix_talib.py
    
from polars_ta.talib import HT_DCPERIOD as ts_HT_DCPERIOD  # noqa
from polars_ta.talib import HT_DCPHASE as ts_HT_DCPHASE  # noqa
from polars_ta.talib import HT_PHASOR as ts_HT_PHASOR  # noqa
from polars_ta.talib import HT_SINE as ts_HT_SINE  # noqa
from polars_ta.talib import HT_TRENDMODE as ts_HT_TRENDMODE  # noqa
from polars_ta.talib import ADD  # noqa
from polars_ta.talib import DIV  # noqa
from polars_ta.talib import MAX  # noqa
from polars_ta.talib import MAXINDEX  # noqa
from polars_ta.talib import MIN  # noqa
from polars_ta.talib import MININDEX  # noqa
from polars_ta.talib import MINMAX  # noqa
from polars_ta.talib import MINMAXINDEX  # noqa
from polars_ta.talib import MULT  # noqa
from polars_ta.talib import SUB  # noqa
from polars_ta.talib import SUM  # noqa
from polars_ta.talib import ACOS  # noqa
from polars_ta.talib import ASIN  # noqa
from polars_ta.talib import ATAN  # noqa
from polars_ta.talib import CEIL  # noqa
from polars_ta.talib import COS  # noqa
from polars_ta.talib import COSH  # noqa
from polars_ta.talib import EXP  # noqa
from polars_ta.talib import FLOOR  # noqa
from polars_ta.talib import LN  # noqa
from polars_ta.talib import LOG10  # noqa
from polars_ta.talib import SIN  # noqa
from polars_ta.talib import SINH  # noqa
from polars_ta.talib import SQRT  # noqa
from polars_ta.talib import TAN  # noqa
from polars_ta.talib import TANH  # noqa
from polars_ta.talib import ADX as ts_ADX  # noqa
from polars_ta.talib import ADXR as ts_ADXR  # noqa
from polars_ta.talib import APO as ts_APO  # noqa
from polars_ta.talib import AROON as ts_AROON  # noqa
from polars_ta.talib import AROONOSC as ts_AROONOSC  # noqa
from polars_ta.talib import BOP as ts_BOP  # noqa
from polars_ta.talib import CCI as ts_CCI  # noqa
from polars_ta.talib import CMO as ts_CMO  # noqa
from polars_ta.talib import DX as ts_DX  # noqa
from polars_ta.talib import MACD as ts_MACD  # noqa
from polars_ta.talib import MACDEXT as ts_MACDEXT  # noqa
from polars_ta.talib import MACDFIX as ts_MACDFIX  # noqa
from polars_ta.talib import MFI as ts_MFI  # noqa
from polars_ta.talib import MINUS_DI as ts_MINUS_DI  # noqa
from polars_ta.talib import MINUS_DM as ts_MINUS_DM  # noqa
from polars_ta.talib import MOM as ts_MOM  # noqa
from polars_ta.talib import PLUS_DI as ts_PLUS_DI  # noqa
from polars_ta.talib import PLUS_DM as ts_PLUS_DM  # noqa
from polars_ta.talib import PPO as ts_PPO  # noqa
from polars_ta.talib import ROC as ts_ROC  # noqa
from polars_ta.talib import ROCP as ts_ROCP  # noqa
from polars_ta.talib import ROCR as ts_ROCR  # noqa
from polars_ta.talib import ROCR100 as ts_ROCR100  # noqa
from polars_ta.talib import RSI as ts_RSI  # noqa
from polars_ta.talib import STOCH as ts_STOCH  # noqa
from polars_ta.talib import STOCHF as ts_STOCHF  # noqa
from polars_ta.talib import STOCHRSI as ts_STOCHRSI  # noqa
from polars_ta.talib import TRIX as ts_TRIX  # noqa
from polars_ta.talib import ULTOSC as ts_ULTOSC  # noqa
from polars_ta.talib import WILLR as ts_WILLR  # noqa
from polars_ta.talib import BBANDS as ts_BBANDS  # noqa
from polars_ta.talib import DEMA as ts_DEMA  # noqa
from polars_ta.talib import EMA as ts_EMA  # noqa
from polars_ta.talib import HT_TRENDLINE as ts_HT_TRENDLINE  # noqa
from polars_ta.talib import KAMA as ts_KAMA  # noqa
from polars_ta.talib import MA as ts_MA  # noqa
from polars_ta.talib import MAMA as ts_MAMA  # noqa
from polars_ta.talib import MAVP as ts_MAVP  # noqa
from polars_ta.talib import MIDPOINT as ts_MIDPOINT  # noqa
from polars_ta.talib import MIDPRICE as ts_MIDPRICE  # noqa
from polars_ta.talib import SAR as ts_SAR  # noqa
from polars_ta.talib import SAREXT as ts_SAREXT  # noqa
from polars_ta.talib import SMA as ts_SMA  # noqa
from polars_ta.talib import T3 as ts_T3  # noqa
from polars_ta.talib import TEMA as ts_TEMA  # noqa
from polars_ta.talib import TRIMA as ts_TRIMA  # noqa
from polars_ta.talib import WMA as ts_WMA  # noqa
from polars_ta.talib import CDL2CROWS as ts_CDL2CROWS  # noqa
from polars_ta.talib import CDL3BLACKCROWS as ts_CDL3BLACKCROWS  # noqa
from polars_ta.talib import CDL3INSIDE as ts_CDL3INSIDE  # noqa
from polars_ta.talib import CDL3LINESTRIKE as ts_CDL3LINESTRIKE  # noqa
from polars_ta.talib import CDL3OUTSIDE as ts_CDL3OUTSIDE  # noqa
from polars_ta.talib import CDL3STARSINSOUTH as ts_CDL3STARSINSOUTH  # noqa
from polars_ta.talib import CDL3WHITESOLDIERS as ts_CDL3WHITESOLDIERS  # noqa
from polars_ta.talib import CDLABANDONEDBABY as ts_CDLABANDONEDBABY  # noqa
from polars_ta.talib import CDLADVANCEBLOCK as ts_CDLADVANCEBLOCK  # noqa
from polars_ta.talib import CDLBELTHOLD as ts_CDLBELTHOLD  # noqa
from polars_ta.talib import CDLBREAKAWAY as ts_CDLBREAKAWAY  # noqa
from polars_ta.talib import CDLCLOSINGMARUBOZU as ts_CDLCLOSINGMARUBOZU  # noqa
from polars_ta.talib import CDLCONCEALBABYSWALL as ts_CDLCONCEALBABYSWALL  # noqa
from polars_ta.talib import CDLCOUNTERATTACK as ts_CDLCOUNTERATTACK  # noqa
from polars_ta.talib import CDLDARKCLOUDCOVER as ts_CDLDARKCLOUDCOVER  # noqa
from polars_ta.talib import CDLDOJI as ts_CDLDOJI  # noqa
from polars_ta.talib import CDLDOJISTAR as ts_CDLDOJISTAR  # noqa
from polars_ta.talib import CDLDRAGONFLYDOJI as ts_CDLDRAGONFLYDOJI  # noqa
from polars_ta.talib import CDLENGULFING as ts_CDLENGULFING  # noqa
from polars_ta.talib import CDLEVENINGDOJISTAR as ts_CDLEVENINGDOJISTAR  # noqa
from polars_ta.talib import CDLEVENINGSTAR as ts_CDLEVENINGSTAR  # noqa
from polars_ta.talib import CDLGAPSIDESIDEWHITE as ts_CDLGAPSIDESIDEWHITE  # noqa
from polars_ta.talib import CDLGRAVESTONEDOJI as ts_CDLGRAVESTONEDOJI  # noqa
from polars_ta.talib import CDLHAMMER as ts_CDLHAMMER  # noqa
from polars_ta.talib import CDLHANGINGMAN as ts_CDLHANGINGMAN  # noqa
from polars_ta.talib import CDLHARAMI as ts_CDLHARAMI  # noqa
from polars_ta.talib import CDLHARAMICROSS as ts_CDLHARAMICROSS  # noqa
from polars_ta.talib import CDLHIGHWAVE as ts_CDLHIGHWAVE  # noqa
from polars_ta.talib import CDLHIKKAKE as ts_CDLHIKKAKE  # noqa
from polars_ta.talib import CDLHIKKAKEMOD as ts_CDLHIKKAKEMOD  # noqa
from polars_ta.talib import CDLHOMINGPIGEON as ts_CDLHOMINGPIGEON  # noqa
from polars_ta.talib import CDLIDENTICAL3CROWS as ts_CDLIDENTICAL3CROWS  # noqa
from polars_ta.talib import CDLINNECK as ts_CDLINNECK  # noqa
from polars_ta.talib import CDLINVERTEDHAMMER as ts_CDLINVERTEDHAMMER  # noqa
from polars_ta.talib import CDLKICKING as ts_CDLKICKING  # noqa
from polars_ta.talib import CDLKICKINGBYLENGTH as ts_CDLKICKINGBYLENGTH  # noqa
from polars_ta.talib import CDLLADDERBOTTOM as ts_CDLLADDERBOTTOM  # noqa
from polars_ta.talib import CDLLONGLEGGEDDOJI as ts_CDLLONGLEGGEDDOJI  # noqa
from polars_ta.talib import CDLLONGLINE as ts_CDLLONGLINE  # noqa
from polars_ta.talib import CDLMARUBOZU as ts_CDLMARUBOZU  # noqa
from polars_ta.talib import CDLMATCHINGLOW as ts_CDLMATCHINGLOW  # noqa
from polars_ta.talib import CDLMATHOLD as ts_CDLMATHOLD  # noqa
from polars_ta.talib import CDLMORNINGDOJISTAR as ts_CDLMORNINGDOJISTAR  # noqa
from polars_ta.talib import CDLMORNINGSTAR as ts_CDLMORNINGSTAR  # noqa
from polars_ta.talib import CDLONNECK as ts_CDLONNECK  # noqa
from polars_ta.talib import CDLPIERCING as ts_CDLPIERCING  # noqa
from polars_ta.talib import CDLRICKSHAWMAN as ts_CDLRICKSHAWMAN  # noqa
from polars_ta.talib import CDLRISEFALL3METHODS as ts_CDLRISEFALL3METHODS  # noqa
from polars_ta.talib import CDLSEPARATINGLINES as ts_CDLSEPARATINGLINES  # noqa
from polars_ta.talib import CDLSHOOTINGSTAR as ts_CDLSHOOTINGSTAR  # noqa
from polars_ta.talib import CDLSHORTLINE as ts_CDLSHORTLINE  # noqa
from polars_ta.talib import CDLSPINNINGTOP as ts_CDLSPINNINGTOP  # noqa
from polars_ta.talib import CDLSTALLEDPATTERN as ts_CDLSTALLEDPATTERN  # noqa
from polars_ta.talib import CDLSTICKSANDWICH as ts_CDLSTICKSANDWICH  # noqa
from polars_ta.talib import CDLTAKURI as ts_CDLTAKURI  # noqa
from polars_ta.talib import CDLTASUKIGAP as ts_CDLTASUKIGAP  # noqa
from polars_ta.talib import CDLTHRUSTING as ts_CDLTHRUSTING  # noqa
from polars_ta.talib import CDLTRISTAR as ts_CDLTRISTAR  # noqa
from polars_ta.talib import CDLUNIQUE3RIVER as ts_CDLUNIQUE3RIVER  # noqa
from polars_ta.talib import CDLUPSIDEGAP2CROWS as ts_CDLUPSIDEGAP2CROWS  # noqa
from polars_ta.talib import CDLXSIDEGAP3METHODS as ts_CDLXSIDEGAP3METHODS  # noqa
from polars_ta.talib import AVGPRICE  # noqa
from polars_ta.talib import MEDPRICE  # noqa
from polars_ta.talib import TYPPRICE  # noqa
from polars_ta.talib import WCLPRICE  # noqa
from polars_ta.talib import BETA as ts_BETA  # noqa
from polars_ta.talib import CORREL as ts_CORREL  # noqa
from polars_ta.talib import LINEARREG as ts_LINEARREG  # noqa
from polars_ta.talib import LINEARREG_ANGLE as ts_LINEARREG_ANGLE  # noqa
from polars_ta.talib import LINEARREG_INTERCEPT as ts_LINEARREG_INTERCEPT  # noqa
from polars_ta.talib import LINEARREG_SLOPE as ts_LINEARREG_SLOPE  # noqa
from polars_ta.talib import STDDEV as ts_STDDEV  # noqa
from polars_ta.talib import TSF as ts_TSF  # noqa
from polars_ta.talib import VAR as ts_VAR  # noqa
from polars_ta.talib import ATR as ts_ATR  # noqa
from polars_ta.talib import NATR as ts_NATR  # noqa
from polars_ta.talib import TRANGE as ts_TRANGE  # noqa
from polars_ta.talib import AD as ts_AD  # noqa
from polars_ta.talib import ADOSC as ts_ADOSC  # noqa
from polars_ta.talib import OBV as ts_OBV  # noqa


from polars import Expr

from polars_ta.ta.price import AVGPRICE
from polars_ta.tdx.reference import MA


def BBI(CLOSE: Expr, M1: int = 3, M2: int = 6, M3: int = 12, M4: int = 20) -> Expr:
    """
    BBI:(MA(CLOSE,M1)+MA(CLOSE,M2)+MA(CLOSE,M3)+MA(CLOSE,M4))/4;

    """
    return AVGPRICE(MA(CLOSE, M1), MA(CLOSE, M2), MA(CLOSE, M3), MA(CLOSE, M4))



from polars import Expr
from polars import Int32, Boolean

from polars_ta import TA_EPSILON
from polars_ta.tdx.reference import SUM


def CROSS(a: Expr, b: Expr) -> Expr:
    return (a <= (b - TA_EPSILON)).shift(1) & (a > (b + TA_EPSILON))


def DOWNNDAY(close: Expr, N: int) -> Expr:
    return NDAY(close.shift(), close, N)


def EVERY(condition: Expr, N: int) -> Expr:
    return SUM(condition.cast(Int32), N) == N


def EXIST(condition: Expr, N: int) -> Expr:
    return SUM(condition.cast(Int32), N) > 0


def EXISTR(condition: Expr, a: int, b: int) -> Expr:
    return EXIST(condition, a - b).shift(b)


def LAST(condition: Expr, a: int, b: int) -> Expr:
    return EVERY(condition, a - b).shift(b)


def LONGCROSS(a: Expr, b: Expr, N: int) -> Expr:
    return CROSS(a, b) & EVERY(a < (b - TA_EPSILON), N).shift(1)


def NDAY(close: Expr, open_: Expr, N: int) -> Expr:
    return EVERY(close > (open_ + TA_EPSILON), N)


def NOT(condition: Expr) -> Expr:
    return ~condition.cast(Boolean)


def UPNDAY(close: Expr, N: int) -> Expr:
    return NDAY(close, close.shift(), N)

# Eastmoney has these two functions
# 东方财富中有此两函数
ALL = EVERY
ANY = EXIST



"""
In this file we use
open_: Expr, high: Expr, low: Expr, close: Expr, high_limit: Expr
open_: Expr, high: Expr, low: Expr, close: Expr, low_limit: Expr
for all parameters
to ensure the similar function signatures when calling them

本文件中参数全用
open_: Expr, high: Expr, low: Expr, close: Expr, high_limit: Expr
open_: Expr, high: Expr, low: Expr, close: Expr, low_limit: Expr
统一的好处是在使用时不用考虑函数调用区别
"""
from polars import Expr

from polars_ta import TA_EPSILON
from polars_ta.candles.cdl1 import four_price_doji, dragonfly, gravestone


def limit_up(price: Expr, high_limit: Expr) -> Expr:
    """
    unified handles for
    reached `high_limit` at `open_`
    reached `high_limit` at `close`
    reached `high_limit` at `high` but `close` < `high_limit`  (at least once reached `high_limit` in this day)
    开盘 涨停
    收盘 涨停
    最高 曾涨停
    """
    return price >= high_limit - TA_EPSILON


def limit_up_at_open(open_: Expr, high: Expr, low: Expr, close: Expr, high_limit: Expr) -> Expr:
    """
    reached `high_limit` at `open_`
    开盘 涨停
    """
    return limit_up(open_, high_limit)


def limit_up_at_close(open_: Expr, high: Expr, low: Expr, close: Expr, high_limit: Expr) -> Expr:
    """
    reached `high_limit` at `close`
    收盘 涨停
    """
    return limit_up(close, high_limit)


def limit_up_at_high(open_: Expr, high: Expr, low: Expr, close: Expr, high_limit: Expr) -> Expr:
    """
    reached `high_limit` at `high` but `close` < `high_limit`  (at least once reached `high_limit` in this day)
    曾涨停
    """
    return limit_up(high, high_limit) & ~limit_up(close, high_limit)


def limit_up_four_price_doji(open_: Expr, high: Expr, low: Expr, close: Expr, high_limit: Expr) -> Expr:
    """
    `low` = `high` = `open_` = `close` = `high_limit`
    一字涨停
    """
    return four_price_doji(open_, high, low, close) & limit_up(close, high_limit)


def limit_up_dragonfly(open_: Expr, high: Expr, low: Expr, close: Expr, high_limit: Expr) -> Expr:
    """
    `low` < `high` = `open_` = `close` = `high_limit`
    T字涨停"""
    return limit_up(close, high_limit) & dragonfly(open_, high, low, close)


# ======================================
def limit_down(price: Expr, low_limit: Expr) -> Expr:
    """
    unified handles for
    reached `low_limit` at `open_`
    reached `low_limit` at `close`
    reached `low_limit` at `high` but `close` > `low_limit`  (at least once reached `low_limit` in this day)
    开盘 跌停
    收盘 跌停
    最低 曾跌停
    """
    return price <= low_limit + TA_EPSILON


def limit_down_at_open(open_: Expr, high: Expr, low: Expr, close: Expr, low_limit: Expr) -> Expr:
    """
    reached `low_limit` at `open_`
    开盘 跌停
    """
    return limit_down(open_, low_limit)


def limit_down_at_close(open_: Expr, high: Expr, low: Expr, close: Expr, low_limit: Expr) -> Expr:
    """
    reached `low_limit` at `close`
    收盘 跌停
    """
    return limit_down(close, low_limit)


def limit_down_at_high(open_: Expr, high: Expr, low: Expr, close: Expr, low_limit: Expr) -> Expr:
    """
    reached `low_limit` at `high` but `close` > `low_limit`  (at least once reached `low_limit` in this day)
    曾跌停
    """
    return limit_down(low, low_limit) & ~limit_down(close, low_limit)


def limit_down_four_price_doji(open_: Expr, high: Expr, low: Expr, close: Expr, low_limit: Expr) -> Expr:
    """
    `low` = `high` = `open_` = `close` = `low_limit`
    一字跌停
    """
    return four_price_doji(open_, high, low, close) & limit_down(close, low_limit)


def limit_down_gravestone(open_: Expr, high: Expr, low: Expr, close: Expr, low_limit: Expr) -> Expr:
    """
    `high` > `low` = `open_` = `close` = `low_limit`
    T字跌停
    """
    return limit_down(open_, low_limit) & gravestone(open_, high, low, close)



"""
In this file we use
open_: Expr, high: Expr, low: Expr, close: Expr
for all parameters
to ensure the similar function signatures when calling them

本文件中参数全用
open_: Expr, high: Expr, low: Expr, close: Expr
统一的好处是在使用时不用考虑函数调用区别
"""
from polars import Expr
from polars import max_horizontal, min_horizontal

from polars_ta import TA_EPSILON


# https://github.com/TA-Lib/ta-lib/blob/main/src/ta_func/ta_utility.h#L327
def real_body(open_: Expr, high: Expr, low: Expr, close: Expr) -> Expr:
    """实体"""
    return (close - open_).abs()


def upper_shadow(open_: Expr, high: Expr, low: Expr, close: Expr) -> Expr:
    """上影"""
    return high - max_horizontal(open_, close)


def lower_shadow(open_: Expr, high: Expr, low: Expr, close: Expr) -> Expr:
    """下影"""
    return min_horizontal(open_, close) - low


def high_low_range(open_: Expr, high: Expr, low: Expr, close: Expr) -> Expr:
    """总长"""
    return high - low


def upper_body(open_: Expr, high: Expr, low: Expr, close: Expr) -> Expr:
    """实体上沿"""
    return max_horizontal(open_, close)


def lower_body(open_: Expr, high: Expr, low: Expr, close: Expr) -> Expr:
    """实体下沿"""
    return min_horizontal(open_, close)


def shadows(open_: Expr, high: Expr, low: Expr, close: Expr) -> Expr:
    """阴影"""
    return high_low_range(open_, high, low, close) - real_body(open_, high, low, close)


def efficiency_ratio(open_: Expr, high: Expr, low: Expr, close: Expr) -> Expr:
    """
    abs(close-open) / (2 * (high-low) - abs(close-open) + TA_EPSILON)
    K线内的市场效率。两个总长减去一个实体长就是路程

    比较粗略的计算市场效率的方法。丢失了部分路程信息，所以结果会偏大
    """
    displacement = real_body(open_, high, low, close)
    distance = 2 * high_low_range(open_, high, low, close) - displacement
    return displacement / (distance + TA_EPSILON)


# ====================================

def candle_color(open_: Expr, high: Expr, low: Expr, close: Expr) -> Expr:
    """K线颜色"""
    return close >= open_


def four_price_doji(open_: Expr, high: Expr, low: Expr, close: Expr) -> Expr:
    """一字"""
    return low >= (high - TA_EPSILON)


def doji(open_: Expr, high: Expr, low: Expr, close: Expr) -> Expr:
    """十字(含一字、T字)"""
    return (open_ - close).abs() <= TA_EPSILON


def dragonfly(open_: Expr, high: Expr, low: Expr, close: Expr) -> Expr:
    """正T字"""
    return doji(open_, high, low, close) & (low < close - TA_EPSILON)


def gravestone(open_: Expr, high: Expr, low: Expr, close: Expr) -> Expr:
    """倒T字"""
    return doji(open_, high, low, close) & (high > close + TA_EPSILON)

# def candle_range(open_: Expr, high: Expr, low: Expr, close: Expr) -> Expr:
#     return real_body(open_, high, low, close)
#     return high_low_range(open_, high, low, close)
#     return shadows(open_, high, low, close)
#
#
# def candle_average(open_: Expr, high: Expr, low: Expr, close: Expr) -> Expr:
#     """平均值"""
#     return high_low_range(open_, high, low, close)



import numpy as np
from numba import jit
from numpy import mean, abs, full, argmax
from numpy.lib.stride_tricks import sliding_window_view


@jit(nopython=True, nogil=True, cache=True)
def roll_avedev(x1, window):
    out = full(x1.shape, np.nan, dtype=float)
    if len(x1) < window:
        return out
    a1 = sliding_window_view(x1, window)
    for i, v1 in enumerate(a1):
        out[i + window - 1] = mean(abs(v1 - mean(v1)))
    return out


@jit(nopython=True, nogil=True, fastmath=True, cache=True)
def roll_bars_since_n(x1, window):
    """BARSSINCEN(X,N): the distance of the first observation that `X != 0` in `N` periods
    BARSSINCEN(X,N):N周期内第一次X不为0到现在的天数

    TODO what if all values are 0?

    TODO 如果一个周期内，都不满足，值取多少？0表当前值满足条件, window-1表示的是区间第0位置的值
    TODO 用window来表示都不满足
    """
    out = full(x1.shape, np.nan, dtype=float)
    if len(x1) < window:
        return out
    a1 = sliding_window_view(x1, window)
    for i, v1 in enumerate(a1):
        p = argmax(v1)
        out[i + window - 1] = window - 1 - p if p or v1[0] else window
    return out



from polars import Boolean
from polars import Expr
from polars import when

from polars_ta.wq.logical import if_else


def IF(condition: Expr, a: Expr, b: Expr) -> Expr:
    """A if X != 0 else B
    IF(X,A,B)若X不为0则返回A,否则返回B
    """
    return if_else(condition.cast(Boolean), a, b)


def IFN(condition: Expr, a: Expr, b: Expr) -> Expr:
    """B if X != 0 else A
    IFN(X,A,B)若X不为0则返回B,否则返回A
    """
    return if_else(condition.cast(Boolean), b, a)


def VALUEWHEN(condition: Expr, x: Expr) -> Expr:
    return when(condition).then(x).otherwise(None).forward_fill()


IFF = IF



from polars import Expr


def ts_max_drawdown(close: Expr) -> Expr:
    """最大回撤"""
    return close - close.cum_max()


def ts_max_drawdown_rate(close: Expr) -> Expr:
    """最大回撤率"""
    return close / close.cum_max() - 1



import numpy as np
from numba import jit, float64, bool_
from numpy import argmax, argmin, prod, mean, std, full, vstack, corrcoef
from numpy.lib.stride_tricks import sliding_window_view


@jit(nopython=True, nogil=True, fastmath=True, cache=True)
def roll_argmax(x1, window, reverse):
    out = full(x1.shape, np.nan, dtype=float)
    if len(x1) < window:
        return out
    a1 = sliding_window_view(x1, window)
    if reverse:
        a1 = a1[:, ::-1]
    for i, v1 in enumerate(a1):
        out[i + window - 1] = argmax(v1)
    return out


@jit(nopython=True, nogil=True, fastmath=True, cache=True)
def roll_argmin(x1, window, reverse):
    out = full(x1.shape, np.nan, dtype=float)
    if len(x1) < window:
        return out
    a1 = sliding_window_view(x1, window)
    if reverse:
        a1 = a1[:, ::-1]
    for i, v1 in enumerate(a1):
        out[i + window - 1] = argmin(v1)
    return out


@jit(nopython=True, nogil=True, fastmath=True, cache=True)
def roll_prod(x1, window):
    out = full(x1.shape, np.nan, dtype=float)
    if len(x1) < window:
        return out
    a1 = sliding_window_view(x1, window)
    for i, v1 in enumerate(a1):
        out[i + window - 1] = prod(v1)
    return out


@jit(nopython=True, nogil=True, fastmath=True, cache=True)
def _co_kurtosis(a1, a2):
    t1 = a1 - mean(a1)
    t2 = a2 - mean(a2)
    t3 = std(a1)
    t4 = std(a2)
    return mean(t1 * (t2 ** 3)) / (t3 * (t4 ** 3))


@jit(nopython=True, nogil=True, fastmath=True, cache=True)
def roll_co_kurtosis(x1, x2, window):
    out = full(x1.shape, np.nan, dtype=float)
    if len(x1) < window:
        return out
    a1 = sliding_window_view(x1, window)
    a2 = sliding_window_view(x2, window)
    for i, (v1, v2) in enumerate(zip(a1, a2)):
        out[i + window - 1] = _co_kurtosis(v1, v2)
    return out


@jit(nopython=True, nogil=True, fastmath=True, cache=True)
def _co_skewness(a1, a2):
    t1 = a1 - mean(a1)
    t2 = a2 - mean(a2)
    t3 = std(a1)
    t4 = std(a2)
    return mean(t1 * (t2 ** 2)) / (t3 * (t4 ** 2))


@jit(nopython=True, nogil=True, fastmath=True, cache=True)
def roll_co_skewness(x1, x2, window):
    out = full(x1.shape, np.nan, dtype=float)
    if len(x1) < window:
        return out
    a1 = sliding_window_view(x1, window)
    a2 = sliding_window_view(x2, window)
    for i, (v1, v2) in enumerate(zip(a1, a2)):
        out[i + window - 1] = _co_skewness(v1, v2)
    return out


@jit(nopython=True, nogil=True, fastmath=True, cache=True)
def _moment(a1, k):
    """centeral moment
    中心矩阵"""
    return mean((a1 - mean(a1)) ** k)


@jit(nopython=True, nogil=True, fastmath=True, cache=True)
def roll_moment(x1, window, k):
    out = full(x1.shape, np.nan, dtype=float)
    if len(x1) < window:
        return out
    a1 = sliding_window_view(x1, window)
    for i, v1 in enumerate(a1):
        out[i + window - 1] = _moment(v1, k)
    return out


@jit(nopython=True, nogil=True, fastmath=True, cache=True)
def _partial_corr(a1, a2, a3):
    """TODO 不知是否正确，需要检查"""
    c = corrcoef(vstack((a1, a2, a3)))
    rxy = c[0, 1]
    rxz = c[0, 2]
    ryz = c[1, 2]
    t1 = rxy - rxz * ryz
    t2 = (1 - rxz ** 2) ** 0.5
    t3 = (1 - ryz ** 2) ** 0.5
    return t1 / (t2 * t3)


@jit(nopython=True, nogil=True, fastmath=True, cache=True)
def roll_partial_corr(x1, x2, x3, window):
    out = full(x1.shape, np.nan, dtype=float)
    if len(x1) < window:
        return out
    a1 = sliding_window_view(x1, window)
    a2 = sliding_window_view(x2, window)
    a3 = sliding_window_view(x3, window)
    for i, (v1, v2, v3) in enumerate(zip(a1, a2, a3)):
        out[i + window - 1] = _partial_corr(v1, v2, v3)
    return out


@jit(nopython=True, nogil=True, fastmath=True, cache=True)
def _triple_corr(a1, a2, a3):
    t1 = a1 - mean(a1)
    t2 = a2 - mean(a2)
    t3 = a3 - mean(a3)
    t4 = std(a1)
    t5 = std(a2)
    t6 = std(a3)
    return mean(t1 * t2 * t3) / (t4 * t5 * t6)


@jit(nopython=True, nogil=True, fastmath=True, cache=True)
def roll_triple_corr(x1, x2, x3, window):
    out = full(x1.shape, np.nan, dtype=float)
    if len(x1) < window:
        return out
    a1 = sliding_window_view(x1, window)
    a2 = sliding_window_view(x2, window)
    a3 = sliding_window_view(x3, window)
    for i, (v1, v2, v3) in enumerate(zip(a1, a2, a3)):
        out[i + window - 1] = _triple_corr(v1, v2, v3)
    return out


@jit(nopython=True, nogil=True, fastmath=True, cache=True)
def isnan(x):
    # https://github.com/numba/numba/issues/2919#issuecomment-747377615
    if int(x) == -9223372036854775808:
        return True
    else:
        return False


@jit(nopython=True, nogil=True, fastmath=True, cache=True)
def _zip_prod(a, b):
    for i in range(1, a.shape[0]):
        if isnan(a[i]):
            a[i] = a[i - 1] * b[i - 1]
    return a


@jit(nopython=True, nogil=True, fastmath=False, cache=False)
def _zip_sum(a, b):
    for i in range(1, a.shape[0]):
        if isnan(a[i]):
            a[i] = a[i - 1] + b[i - 1]
    return a


@jit(float64[:](bool_[:], bool_[:], bool_[:], bool_[:], bool_, bool_), nopython=True, fastmath=True, nogil=True, cache=True)
def signals_to_amount(is_long_entry: np.ndarray, is_long_exit: np.ndarray,
                      is_short_entry: np.ndarray, is_short_exit: np.ndarray,
                      accumulate: bool = False,
                      action: bool = False) -> np.ndarray:
    """将4路信号转换成持仓状态。适合按资产分组后的长表

    在`LongOnly`场景下，`is_short_entry`和`is_short_exit`输入数据值都为`False`即可

    Parameters
    ----------
    is_long_entry: np.ndarray
        是否多头入场
    is_long_exit: np.ndarray
        是否多头出场
    is_short_entry: np.ndarray
        是否空头入场
    is_short_exit: np.ndarray
        是否空头出场
    accumulate: bool
        遇到重复信号时是否累计
    action: bool
        返回持仓状态还是下单操作

    Returns
    -------
    np.ndarray
        持仓状态

    Examples
    --------
    ```python
    long_entry = np.array([True, True, False, False, False])
    long_exit = np.array([False, False, True, False, False])
    short_entry = np.array([False, False, True, False, False])
    short_exit = np.array([False, False, False, True, False])

    amount = signals_to_amount(long_entry, long_exit, short_entry, short_exit, accumulate=True, action=False)
    ```

    """
    _amount: float = 0.0  # 持仓状态
    _action: float = 0.0  # 下单方向
    out = np.zeros(len(is_long_entry), dtype=float)
    for i in range(len(is_long_entry)):
        if _amount == 0.0:
            # 多头信号优先级高于空头信号
            if is_long_entry[i]:
                _amount += 1.0
                _action = 1.0
            elif is_short_entry[i]:
                _amount -= 1.0
                _action = -1.0
        elif _amount > 0.0:
            if is_long_exit[i]:
                _amount -= 1.0
                _action = -1.0
            elif is_long_entry[i] and accumulate:
                _amount += 1.0
                _action = 1.0
        else:
            if is_short_exit[i]:
                _amount += 1.0
                _action = 1.0
            elif is_short_entry[i] and accumulate:
                _amount -= 1.0
                _action = -1.0

        out[i] = _action if action else _amount
    return out



from polars import Expr

from polars_ta.wq.time_series import ts_arg_max
from polars_ta.wq.time_series import ts_arg_min
from polars_ta.wq.time_series import ts_max
from polars_ta.wq.time_series import ts_min
from polars_ta.wq.time_series import ts_sum


def ADD(high: Expr, low: Expr) -> Expr:
    return high + low


def DIV(high: Expr, low: Expr) -> Expr:
    return high / low


def MAX(close: Expr, timeperiod: int = 30) -> Expr:
    """

    Notes
    -----
    It is the maximum value of the time series, not the maximum value of multiple columns (max_horizontal)
    时序上窗口最大，不要与多列最大搞混

    """
    return ts_max(close, timeperiod)


def MAXINDEX(close: Expr, timeperiod: int = 30) -> Expr:
    """
    Comparing to `ts_arg_max` this also marks the abs. position of the max value
    与ts_arg_max的区别是，标记了每个区间最大值的绝对位置，可用来画图标记
    """
    a = close.cum_count()
    b = ts_arg_max(close, timeperiod)
    return a - b - 1


def MIN(close: Expr, timeperiod: int = 30) -> Expr:
    return ts_min(close, timeperiod)


def MININDEX(close: Expr, timeperiod: int = 30) -> Expr:
    a = close.cum_count()
    b = ts_arg_min(close, timeperiod)
    return a - b - 1


def MULT(high: Expr, low: Expr) -> Expr:
    return high * low


def SUB(high: Expr, low: Expr) -> Expr:
    return high - low


def SUM(close: Expr, timeperiod: int = 30) -> Expr:
    return ts_sum(close, timeperiod)



from polars import Expr, all_horizontal, any_horizontal
from polars import when, Int8, Boolean


def and_(a: Expr, b: Expr, *args) -> Expr:
    """Logical AND operator, returns true if both operands are true and returns false otherwise"""
    return all_horizontal(a, b, *args)


def equal(input1: Expr, input2: Expr) -> Expr:
    """Returns true if both inputs are same and returns false otherwise"""
    return input1 == input2


def if_else(input1: Expr, input2: Expr, input3: Expr = None) -> Expr:
    """If input1 is true then return input2 else return input3."""
    return when(input1).then(input2).otherwise(input3)


def is_finite(input1: Expr) -> Expr:
    """If (input NaN or input == INF) return 0, else return 1"""
    return input1.is_finite()


def is_nan(input1: Expr) -> Expr:
    """If (input == NaN) return 1 else return 0"""
    return input1.is_nan()


def is_not_finite(input1: Expr) -> Expr:
    """If (input NAN or input == INF) return 1 else return 0"""
    return input1.is_infinite()


def is_not_nan(input1: Expr) -> Expr:
    """If (input != NaN) return 1 else return 0"""
    return input1.is_not_nan()


def less(input1: Expr, input2: Expr) -> Expr:
    """If input1 < input2 return true, else return false"""
    return input1 < input2


def negate(input1: Expr) -> Expr:
    """The result is true if the converted operand is false; the result is false if the converted operand is true"""
    return ~input1


def or_(a: Expr, b: Expr, *args) -> Expr:
    """Logical OR operator returns true if either or both inputs are true and returns false otherwise"""
    return any_horizontal(a, b, *args)


def xor(a: Expr, b: Expr) -> Expr:
    """Logical XOR operator returns true if exactly one of the inputs is true and returns false otherwise"""
    return a.xor(b)


def int_(a: Expr) -> Expr:
    """convert bool to int"""
    return a.cast(Int8)


def bool_(a: Expr) -> Expr:
    """convert int to bool"""
    return a.cast(Boolean)



# this code is auto generated by tools/prefix_tdx.py

from polars_ta.tdx.arithmetic import ABS  # noqa
from polars_ta.tdx.arithmetic import ACOS  # noqa
from polars_ta.tdx.arithmetic import ADD  # noqa
from polars_ta.tdx.arithmetic import ASIN  # noqa
from polars_ta.tdx.arithmetic import ATAN  # noqa
from polars_ta.tdx.arithmetic import BETWEEN  # noqa
from polars_ta.tdx.arithmetic import CEILING  # noqa
from polars_ta.tdx.arithmetic import COS  # noqa
from polars_ta.tdx.arithmetic import EXP  # noqa
from polars_ta.tdx.arithmetic import FLOOR  # noqa
from polars_ta.tdx.arithmetic import FRACPART  # noqa
from polars_ta.tdx.arithmetic import INTPART  # noqa
from polars_ta.tdx.arithmetic import LN  # noqa
from polars_ta.tdx.arithmetic import LOG  # noqa
from polars_ta.tdx.arithmetic import MAX  # noqa
from polars_ta.tdx.arithmetic import MIN  # noqa
from polars_ta.tdx.arithmetic import MOD  # noqa
from polars_ta.tdx.arithmetic import POW  # noqa
from polars_ta.tdx.arithmetic import REVERSE  # noqa
from polars_ta.tdx.arithmetic import ROUND  # noqa
from polars_ta.tdx.arithmetic import ROUND2  # noqa
from polars_ta.tdx.arithmetic import SGN  # noqa
from polars_ta.tdx.arithmetic import SIGN  # noqa
from polars_ta.tdx.arithmetic import SIN  # noqa
from polars_ta.tdx.arithmetic import SQRT  # noqa
from polars_ta.tdx.arithmetic import SUB  # noqa
from polars_ta.tdx.arithmetic import TAN  # noqa
from polars_ta.tdx.choice import IF  # noqa
from polars_ta.tdx.choice import IFF  # noqa
from polars_ta.tdx.choice import IFN  # noqa
from polars_ta.tdx.choice import VALUEWHEN  # noqa
from polars_ta.tdx.energy import BRAR_AR as ts_BRAR_AR  # noqa
from polars_ta.tdx.energy import BRAR_BR as ts_BRAR_BR  # noqa
from polars_ta.tdx.energy import CR as ts_CR  # noqa
from polars_ta.tdx.energy import MASS as ts_MASS  # noqa
from polars_ta.tdx.energy import PSY as ts_PSY  # noqa
from polars_ta.tdx.logical import ALL as ts_ALL  # noqa
from polars_ta.tdx.logical import ANY as ts_ANY  # noqa
from polars_ta.tdx.logical import CROSS as ts_CROSS  # noqa
from polars_ta.tdx.logical import DOWNNDAY as ts_DOWNNDAY  # noqa
from polars_ta.tdx.logical import EVERY as ts_EVERY  # noqa
from polars_ta.tdx.logical import EXIST as ts_EXIST  # noqa
from polars_ta.tdx.logical import EXISTR as ts_EXISTR  # noqa
from polars_ta.tdx.logical import LAST as ts_LAST  # noqa
from polars_ta.tdx.logical import LONGCROSS as ts_LONGCROSS  # noqa
from polars_ta.tdx.logical import NDAY as ts_NDAY  # noqa
from polars_ta.tdx.logical import NOT  # noqa
from polars_ta.tdx.logical import UPNDAY as ts_UPNDAY  # noqa
from polars_ta.tdx.moving_average import BBI as ts_BBI  # noqa
from polars_ta.tdx.over_bought_over_sold import ATR as ts_ATR  # noqa
from polars_ta.tdx.over_bought_over_sold import BIAS as ts_BIAS  # noqa
from polars_ta.tdx.over_bought_over_sold import CCI as ts_CCI  # noqa
from polars_ta.tdx.over_bought_over_sold import KDJ as ts_KDJ  # noqa
from polars_ta.tdx.over_bought_over_sold import MFI as ts_MFI  # noqa
from polars_ta.tdx.over_bought_over_sold import MTM as ts_MTM  # noqa
from polars_ta.tdx.over_bought_over_sold import RSI as ts_RSI  # noqa
from polars_ta.tdx.over_bought_over_sold import RSV as ts_RSV  # noqa
from polars_ta.tdx.pressure_support import BOLL as ts_BOLL  # noqa
from polars_ta.tdx.reference import BARSLAST as ts_BARSLAST  # noqa
from polars_ta.tdx.reference import BARSLASTCOUNT as ts_BARSLASTCOUNT  # noqa
from polars_ta.tdx.reference import BARSSINCE as ts_BARSSINCE  # noqa
from polars_ta.tdx.reference import BARSSINCEN as ts_BARSSINCEN  # noqa
from polars_ta.tdx.reference import CUMSUM as ts_CUMSUM  # noqa
from polars_ta.tdx.reference import DMA as ts_DMA  # noqa
from polars_ta.tdx.reference import EMA as ts_EMA  # noqa
from polars_ta.tdx.reference import EXPMA as ts_EXPMA  # noqa
from polars_ta.tdx.reference import EXPMEMA as ts_EXPMEMA  # noqa
from polars_ta.tdx.reference import FILTER as ts_FILTER  # noqa
from polars_ta.tdx.reference import HOD as ts_HOD  # noqa
from polars_ta.tdx.reference import LOD as ts_LOD  # noqa
from polars_ta.tdx.reference import MAX  # noqa
from polars_ta.tdx.reference import MEMA as ts_MEMA  # noqa
from polars_ta.tdx.reference import MIN  # noqa
from polars_ta.tdx.reference import RANGE  # noqa
from polars_ta.tdx.reference import SMA_CN as ts_SMA_CN  # noqa
from polars_ta.tdx.reference import SUMIF as ts_SUMIF  # noqa
from polars_ta.tdx.reference import TMA as ts_TMA  # noqa
from polars_ta.tdx.reference import TR as ts_TR  # noqa
from polars_ta.tdx.statistic import AVEDEV as ts_AVEDEV  # noqa
from polars_ta.tdx.statistic import DEVSQ as ts_DEVSQ  # noqa
from polars_ta.tdx.statistic import SLOPE as ts_SLOPE  # noqa
from polars_ta.tdx.statistic import STD as ts_STD  # noqa
from polars_ta.tdx.statistic import STDDEV as ts_STDDEV  # noqa
from polars_ta.tdx.statistic import STDP as ts_STDP  # noqa
from polars_ta.tdx.statistic import VAR as ts_VAR  # noqa
from polars_ta.tdx.statistic import VARP as ts_VARP  # noqa
from polars_ta.tdx.trend import ADX as ts_ADX  # noqa
from polars_ta.tdx.trend import ADXR as ts_ADXR  # noqa
from polars_ta.tdx.trend import DPO as ts_DPO  # noqa
from polars_ta.tdx.trend import EMV as ts_EMV  # noqa
from polars_ta.tdx.trend import MINUS_DI as ts_MINUS_DI  # noqa
from polars_ta.tdx.trend import MINUS_DM as ts_MINUS_DM  # noqa
from polars_ta.tdx.trend import PLUS_DI as ts_PLUS_DI  # noqa
from polars_ta.tdx.trend import PLUS_DM as ts_PLUS_DM  # noqa
from polars_ta.tdx.volume import OBV as ts_OBV  # noqa
from polars_ta.tdx.volume import VR as ts_VR  # noqa


import polars as pl
from polars import selectors as cs


def ts_pit(df: pl.DataFrame, func=None,
           date='date', update_time='update_time', asset='asset'):
    """Computing Point In Time

    Parameters
    ----------
    df: pl.DataFrame
        dataframe after group_by(date)
    func:
        apply to the dataframe output from pit, can be None
    date:str
        column name, group by this col
    update_time:str
        column name, group by this col
    asset:str
        asset column name
        资产。asfreq时，asset为空，后面要分组时不方便

    Returns
    -------
    pd.DataFrame

    Notes
    -----
    1. Data update might be weekeng
    2. You cannot use this for further time-series computation.
        Alert for inclusion of future data
    3. This is different from general `ts_` functions, be careful


    Point In Time计算

    Parameters
    ----------
    df: pl.DataFrame
        group_by(date)后的数据块
    func:
        作用于pit提取出的DataFrame，可不填
    date:str
        分组依据
    update_time:str
        分组依据
    asset:str
        资产。asfreq时，asset为空，后面要分组时不方便

    Returns
    -------
    pd.DataFrame

    Notes
    -----
    1. 数据更新可能是周末
    2. 输出的数据不能再做时序计算，因为再算就引入未来数据了
    3. 此算子与一般的`ts_`算子不同，小心使用

    """
    # sort first
    # 一定要提前排序
    sort_by = [date, update_time]
    df = df.sort(sort_by)

    dr = (
        # group by, record the count and update time
        # 分组，记录数量，和更新时间
        df.group_by(date).agg(update_time, pl.len())
        # 多加一行是否最后值
        .filter(pl.col('len') > 1)
        .select(pl.col(update_time).list.slice(1))
        .explode(update_time).to_series()
    )

    # 最大的更新时间
    max_update_time = df.select(update_time).max().to_series()
    dr = dr.append(max_update_time).unique().sort().to_list()

    # fill missing data to 4 quarters
    # 部分很早的财报有缺失，比如只有年报或半年报，仿asfreq补充成4季
    min_max_q = df.select(date).unique().upsample(date, every='1q')
    # TODO 使用date填充update_time是否会引入问题？
    df = df.join(min_max_q, on=date, how='outer_coalesce').with_columns(pl.col(update_time).fill_null(pl.col(date)))
    # 由于asfreq导致asset可能为空，但之后可能要用到，所以填充一下
    df = df.with_columns(pl.col(asset).forward_fill())

    dd = []
    # iterate over the key dates
    # 遍历关键日期
    for dt in dr:
        # 每次只看指定更新日之间的记录
        d: pl.DataFrame = df.filter(pl.col(update_time) <= dt)
        # 已经排序了，只取能取最新值
        d = d.group_by(date, maintain_order=True).last()
        # TODO 其实asfreq放这更合理，但为了提速将其提前

        d = d.with_columns(pl.col(date).dt.quarter().alias('quarter'))
        if func is not None:
            # 分块计算
            d = func(d)
        dd.append(d)

    # sort by the date and update time
    # 按报告期排序
    x = pl.concat(dd)
    x = x.unique(subset=sort_by, keep='first').sort(sort_by)

    return x


def period_to_quarter():
    """时段数据转单成季"""
    csnum = cs.numeric().exclude('quarter')
    return pl.when(pl.col('quarter') == 1).then(csnum).otherwise(csnum.diff()).name.suffix('_Q')


def peroid_to_ttm():
    """时段数据计算ttm"""
    csnum = cs.numeric().exclude('quarter')
    # 年报数据
    f1 = pl.when(pl.col('quarter') == 4).then(csnum).otherwise(None)
    # 上年年报和同比都存在。当前报告期-上年报告期+上年年报
    f2 = csnum - csnum.shift(4) + f1.forward_fill()
    # 年化计算法=当前报告期*年化系数
    f3 = csnum / pl.col('quarter') * 4
    # return pl.coalesce(f1, f2, f3).name.suffix('_TTM')
    return (pl.when(f1.is_not_null()).then(f1)
            .when(f2.is_not_null()).then(f2)
            .when(f3.is_not_null()).then(f3)
            .otherwise(None)).name.suffix('_TTM')


def point_to_ttm() -> pl.Expr:
    """时点数据计算ttm"""
    csnum = cs.numeric().exclude('quarter')
    return csnum.rolling_mean(4).name.suffix('_TTM')



from polars_ta.tdx.arithmetic import *  # noqa
from polars_ta.tdx.choice import *  # noqa
from polars_ta.tdx.energy import *  # noqa
from polars_ta.tdx.logical import *  # noqa
from polars_ta.tdx.moving_average import *  # noqa
from polars_ta.tdx.over_bought_over_sold import *  # noqa
from polars_ta.tdx.pressure_support import *  # noqa
from polars_ta.tdx.reference import *  # noqa
from polars_ta.tdx.statistic import *  # noqa
from polars_ta.tdx.trend import *  # noqa
from polars_ta.tdx.volume import *  # noqa



from polars import Expr

from polars_ta import TA_EPSILON
from polars_ta.ta.overlap import EMA


def AD(high: Expr, low: Expr, close: Expr, volume: Expr) -> Expr:
    ad = ((close - low) - (high - close)) / (high - low + TA_EPSILON) * volume
    return ad.cum_sum()


def ADOSC(high: Expr, low: Expr, close: Expr, volume: Expr, fastperiod: int = 3, slowperiod: int = 10) -> Expr:
    ad = AD(high, low, close, volume)
    return EMA(ad, fastperiod) - EMA(ad, slowperiod)


def OBV(close: Expr, volume: Expr) -> Expr:
    """"""
    # using volume for the first value will be exactly the same as talib.OBV
    # 第一个值用volume就与talib.OBV完全一样了
    obv = close.diff().sign().fill_null(1) * volume
    return obv.cum_sum()
